
/* GOSSIP - The Global Ops admin tool thingy */

#define GOSSIP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>
#include <config.h>
#include <gossip.h>
#include <logger.h>
#include <version.h>
#include <queue.h>
#include <configs.h>
#include <misc.h>
#include <server.h>
#include <client.h>

int main(int argc, char **argv) {
   struct sockaddr_in;
   int c;

   while ((c = getopt(argc, argv, "dDvhl:L:")) != -1) { 
      switch (c) {
         case 'd' :
            do_daemonize=1;
            break;
	 case 'D' : 
	    do_debug=1;
	    break;
         case 'v':
            version();
            break;
         case 'l':
            logfile=optarg;
            break;
         case 'L':
            loggingmethod=atoi(optarg);
            if (loggingmethod <0 || loggingmethod >3)
               usage(argv[0]);
            break;
         case '?':
         case 'h':
         default:
            usage(argv[0]);
            break;
	}
   }
   

   /* Init stuff */
   init();

      
   /* Main loop */
   while(1) {
      while (!connected()) {
         if (connecttogo()<0) sleep(CONNECT_RETRY_WAITTIME);
      }
     
      /* Are there any clients trying to connect? Accept them or */
      /* disconnect them if GOSSIP is full 			 */ 
      handle_incoming();
      
      /* Handle server-data */
      handle_serverdata();
      
      /* Handle client-data */
      handle_clientdata();
                      
      /* Handle log-relaying */
      if (do_logrelay==1) handle_logrelay();
      
      /* Handle timeouts */
      handle_timeouts();

  } /* Main loop */
      
   exit(1);
}

void init(void) {
   int i;
   struct itimerval itv;
      
   /* Initialize logger */
   init_logger();

   /* Initialize queue */
   init_queue();
   
   /* Parse config file */
   do_config();
   
   if (listenport==0 || adminport==0 || strlen(adminpass)==0 || connects_per_ip<1) {
      log(LOGGER_ERROR,"Configuration invalid, please fix");
      exit(1);
   }

   /* Set sockets to inactive */
   memset(sockets,-1,sizeof(sockets));
 
   /* Initialize client-info  */
   for (i=0;i<MAXADMINS;i++) {
      clients[i].socket=-1;
   }

   if (do_listen()) {
      log(LOGGER_ERROR,"Starting listener failed, gossip exiting");
      exit(1);
   }
   
   if (do_daemonize) {
      if (daemonize()) {
         perror("Could not daemonize");
         exit(1);
      }
   }

   /* Turn on logrelay and turn logfile checking off */
   do_logrelay=1;
   do_checklogfiles=0;

   /* Set signal handler */
   signal(SIGALRM, handle_alarm);
   
   /* Set alarm */
   itv.it_interval.tv_usec=0;
   itv.it_interval.tv_sec=1;
   itv.it_value.tv_usec=0;
   itv.it_value.tv_sec=1;
   setitimer(ITIMER_REAL,&itv,NULL);
  
}


/* Every now and then a alarm-signal is sent and this function is called */
/* By setting do_checklogfiles to 1, another function will check if      */
/* there's a new logfile to read 					 */ 
void handle_alarm(int sig) {
   do_checklogfiles=1;
}


/* Start a listener for the clients */
int do_listen(void) {
   struct sockaddr_in my_addr;
   int yes=1;
   
   if((sockets[1]=socket(AF_INET,SOCK_STREAM,0))==-1) {
      log(LOGGER_ERROR,"Could not create listen socket: %s",strerror(errno));
      return -1;
   }
   
   /* Make address reusable */
   setsockopt(sockets[1],SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(yes));

   /* Set socket to non-blocking */
   fcntl(sockets[1],F_SETFL,O_NONBLOCK);
   
   my_addr.sin_family=AF_INET;
   my_addr.sin_port=htons(listenport);
   my_addr.sin_addr.s_addr=INADDR_ANY;
   memset(&(my_addr.sin_zero),'\0',8); /* zero rest of struct */
   
   if (bind(sockets[1],(struct sockaddr *)&my_addr,sizeof(struct sockaddr))==-1) {
      log(LOGGER_ERROR,"Could not bind to port %d: %s",listenport,strerror(errno));
      return -1;
   }   
 
   if (listen(sockets[1],MAXADMINS)==-1) {
      log(LOGGER_ERROR,"Could not listen: %s",strerror(errno));
      return -1;
   }   
   
   log(LOGGER_DEBUG,"Listening on port %d",listenport);   
   return 0;
}
                   

/* Checks if the buffer contains a gossip-command and handles it 	*/
/* Returns true if it was a GOSSIP-command, false if not		*/
int handle_gossipcommand(int client, char *buf, int bufsize) {
   char tempstr[64];
   int retval=0;
      
   if (bufsize>sizeof(tempstr)-1) return 0; /* It's too big, can't be a GOSSIP-command */

   strncpy(tempstr,buf,bufsize);

   if (!strcmp(tempstr,"GOSSIP|start")) {
      log(LOGGER_DEBUG,"Client %d (%s) enabled log-relaying",client,clients[client].ip);
      clients[client].relaylogs=1;
      retval=1;
   }
   if (!strcmp(tempstr,"GOSSIP|stop")) {
      log(LOGGER_DEBUG,"Client %d (%s) disabled log-relaying",client,clients[client].ip);
      clients[client].relaylogs=0;
      retval=1;
   }
   return retval;
}


/* Relay GO-logs to the clients */
/* Could use some cleaning-up	*/
void handle_logrelay(void) {
   static char currentlogfile[512]="\0";
   static FILE *fp;
   static long position;
         
   char templogfile[512]; 
   struct stat dirstat;
   DIR *dir;
   struct dirent *direntry;
   struct dirent *tempdirentry;
   char readbuf[BUFSIZ];
   char writebuf[BUFSIZ+16]; /* To hold the readbuffer prepended with "GOSSIP|" */
      

   /* Check logfiles every now and then, if do_checklogfiles is set */
   if (do_checklogfiles) {
      do_checklogfiles=0;
      /* Open directory */
      if ((dir=opendir(logdirectory))==NULL) {
         log(LOGGER_WARNING, "Could not open logdirectory %s, logrelaying disabled",logdirectory);
         do_logrelay=0;
         return;
      } 
   
      /* Find last entry in the directory */
      direntry=readdir(dir);
      if (direntry==NULL) {
         /* No logfiles in directory */
         return;
      }

      while ((tempdirentry=readdir(dir))!=NULL) direntry=tempdirentry; 			/* Get last entry */
      snprintf(templogfile,sizeof(templogfile),"%s/%s",logdirectory,direntry->d_name);  /* Build full filename */
 
      /* Close directory */
      closedir(dir);
   
      /* Check if it's a file */ 
      if (lstat(templogfile,&dirstat)) {
        log(LOGGER_WARNING,"Could not stat logfile %s: %s",templogfile,strerror(errno));
        closedir(dir);
        return;
      } else {
         if (!(dirstat.st_mode && S_IFREG)) {
	    /* Logfile is not a file, might be dangerous to read it, so don't */
            closedir(dir);
            return;
         }
      }
   
      /* Check if it's a new logfile, if so, close old and open new */ 
      if (strcmp(currentlogfile,templogfile)) {
         if (strlen(currentlogfile)>0) fclose(fp);
      
         strncpy(currentlogfile,templogfile,sizeof(currentlogfile));
         log(LOGGER_DEBUG, "Now using logfile %s",currentlogfile);

         if ((fp=fopen(currentlogfile,"r"))==NULL) {
            log(LOGGER_WARNING,"Could not open logfile %s: %s", currentlogfile, strerror(errno));
            return;
         }
         fseek(fp,0,SEEK_END); /* Start at the end of the file, IS THIS OK? */
         position=(long)ftell(fp);
      }            
   }
   

   if (strlen(currentlogfile)==0) return; /* If logfile isn't set, back out */

   /* Read to the end of the file and relay the read lines */
   if (position<(long)dirstat.st_size) {
      while(!feof(fp)) {
         if(fgets(readbuf,sizeof(readbuf),fp)==0) break;
     
         if (ferror(fp)) {
  	    log(LOGGER_WARNING,"Error reading logfile %s: %s", currentlogfile, strerror(errno));        
  	    fclose(fp);
 	    currentlogfile[0]='\0';
 	    return;
         }
      
         /* Relay data */
         /* Put "GOSSIP|" in front, then broadcast it to the clients */
         snprintf(writebuf,sizeof(writebuf),"GOSSIP|%s",readbuf);
         do_broadcast(writebuf,strlen(writebuf));      
      }
   }
   position=ftell(fp);  /* Save position for next time */
   clearerr(fp); 	/* Clear EOF status */
}


/* Guess what this function does */
void usage(char *pgm) {
        printf("Usage: %s [-dvh] [-l logfile] [-L logmethod]\n",pgm);
        printf("Options:\n");
        printf("\t-d\t\tDaemonize, default is to not deamonize\n");
	printf("\t-D\t\tDebug, default is not to show debug-information\n");
        printf("\t-v\t\tShow version\n");
        printf("\t-h\t\tShow usage (this text)\n");
        printf("\t-l [Logfile]\tUse the given filename as logfile\n");
        printf("\t-L [Logmethod]\tLog using the given method:\n");
        printf("\t\t\t0 - Don't log\n");
        printf("\t\t\t1 - Log to syslog only\n");
        printf("\t\t\t2 - Log to file only\n");
        printf("\t\t\t3 - Log to both syslog and file\n");
        printf("\n");
        exit(1);
}

/* EOF */     


              